# Unit tests for structural_navigator.py methods.
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position
# pylint: disable=import-outside-toplevel
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-statements
# pylint: disable=protected-access
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-locals
# pylint: disable=too-many-lines

"""Unit tests for structural_navigator.py methods."""

from __future__ import annotations

from typing import TYPE_CHECKING

import gi
import pytest

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

if TYPE_CHECKING:
    from .orca_test_context import OrcaTestContext


@pytest.mark.unit
class TestStructuralNavigator:
    """Test StructuralNavigator class methods."""

    def _setup_dependencies(self, test_context):
        """Set up dependencies for structural_navigator module testing."""

        additional_modules = [
            "orca.cmdnames",
            "orca.dbus_service",
            "orca.debug",
            "orca.focus_manager",
            "orca.guilabels",
            "orca.input_event_manager",
            "orca.keybindings",
            "orca.messages",
            "orca.object_properties",
            "orca.orca_gui_navlist",
            "orca.orca_i18n",
            "orca.script_manager",
            "orca.settings_manager",
            "orca.AXHypertext",
            "orca.AXObject",
            "orca.AXTable",
            "orca.AXText",
            "orca.AXUtilities",
            "orca.input_event",
        ]
        essential_modules = test_context.setup_shared_dependencies(additional_modules)
        self._setup_mocks(test_context, essential_modules)
        return essential_modules

    def _setup_cycle_navigation_mode_mocks(
        self, test_context, nav, current_mode, supports_collection=True
    ):
        """Set up common navigator mocks for cycle_mode testing scenarios."""

        mock_get_mode = test_context.Mock(return_value=current_mode)
        test_context.patch_object(nav, "get_mode", new=mock_get_mode)

        mock_set_mode = test_context.Mock()
        test_context.patch_object(nav, "set_mode", new=mock_set_mode)

        test_context.patch_object(nav, "_is_active_script", return_value=True)

        mock_determine_root = None
        if supports_collection:
            mock_root = test_context.Mock()
            mock_determine_root = test_context.Mock(return_value=mock_root)
            test_context.patch_object(nav, "_determine_root_container", new=mock_determine_root)

        return mock_get_mode, mock_set_mode, mock_determine_root

    def _setup_mocks(self, test_context, essential_modules) -> None:
        """Set up common mocks for StructuralNavigator tests."""

        input_event_handler_mock = test_context.Mock()
        essential_modules["orca.input_event"].InputEventHandler = test_context.Mock(
            return_value=input_event_handler_mock
        )
        key_bindings_instance = test_context.Mock()
        essential_modules["orca.keybindings"].KeyBindings = test_context.Mock(
            return_value=key_bindings_instance
        )
        settings_manager_instance = test_context.Mock()
        settings_manager_instance.get_setting.return_value = True
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value = settings_manager_instance
        controller_mock = test_context.Mock()
        controller_mock.register_decorated_module.return_value = None
        essential_modules["orca.dbus_service"].get_remote_controller.return_value = controller_mock
        essential_modules["orca.debug"].print_message = test_context.Mock()
        essential_modules["orca.debug"].LEVEL_INFO = 800
        focus_manager_instance = test_context.Mock()
        focus_manager_instance.get_locus_of_focus.return_value = None
        essential_modules["orca.focus_manager"].get_manager.return_value = focus_manager_instance
        essential_modules["orca.AXObject"].get_parent.return_value = None
        essential_modules["orca.AXObject"].supports_collection.return_value = True
        essential_modules["orca.AXObject"].find_ancestor.return_value = None
        essential_modules["orca.AXUtilities"].is_web_application.return_value = False
        essential_modules["orca.AXUtilities"].is_heading.return_value = False
        essential_modules["orca.AXUtilities"].find_descendant.return_value = None
        script_manager_instance = test_context.Mock()
        script_manager_instance.get_active_script.return_value = None
        essential_modules["orca.script_manager"].get_manager.return_value = script_manager_instance
        essential_modules["orca.orca_i18n"]._ = lambda msg: msg
        essential_modules["orca.messages"].STRUCTURAL_NAVIGATION_KEYS_DOCUMENT = "Document mode"
        essential_modules["orca.messages"].STRUCTURAL_NAVIGATION_KEYS_GUI = "GUI mode"
        essential_modules["orca.messages"].STRUCTURAL_NAVIGATION_KEYS_OFF = "Off mode"
        essential_modules[
            "orca.messages"
        ].STRUCTURAL_NAVIGATION_NOT_SUPPORTED_FULL = "Not supported full"
        essential_modules[
            "orca.messages"
        ].STRUCTURAL_NAVIGATION_NOT_SUPPORTED_BRIEF = "Not supported brief"
        essential_modules["orca.messages"].WRAPPING_TO_TOP = "Wrapping to top"
        essential_modules["orca.messages"].WRAPPING_TO_BOTTOM = "Wrapping to bottom"
        essential_modules["orca.cmdnames"].STRUCTURAL_NAVIGATION_MODE_CYCLE = "cycle_mode"
        essential_modules["orca.cmdnames"].BLOCKQUOTE_PREV = "previous_blockquote"
        essential_modules["orca.cmdnames"].BLOCKQUOTE_NEXT = "next_blockquote"
        essential_modules["orca.cmdnames"].BLOCKQUOTE_LIST = "list_blockquotes"
        essential_modules["orca.cmdnames"].BUTTON_PREV = "previous_button"
        essential_modules["orca.cmdnames"].BUTTON_NEXT = "next_button"
        essential_modules["orca.cmdnames"].BUTTON_LIST = "list_buttons"
        essential_modules["orca.cmdnames"].CHECK_BOX_PREV = "previous_checkbox"
        essential_modules["orca.cmdnames"].CHECK_BOX_NEXT = "next_checkbox"
        essential_modules["orca.cmdnames"].CHECK_BOX_LIST = "list_checkboxes"
        essential_modules["orca.cmdnames"].COMBO_BOX_PREV = "previous_combobox"
        essential_modules["orca.cmdnames"].COMBO_BOX_NEXT = "next_combobox"
        essential_modules["orca.cmdnames"].COMBO_BOX_LIST = "list_comboboxes"
        essential_modules["orca.cmdnames"].ENTRY_PREV = "previous_entry"
        essential_modules["orca.cmdnames"].ENTRY_NEXT = "next_entry"
        essential_modules["orca.cmdnames"].ENTRY_LIST = "list_entries"
        essential_modules["orca.cmdnames"].FORM_FIELD_PREV = "previous_form_field"
        essential_modules["orca.cmdnames"].FORM_FIELD_NEXT = "next_form_field"
        essential_modules["orca.cmdnames"].FORM_FIELD_LIST = "list_form_fields"
        essential_modules["orca.cmdnames"].HEADING_PREV = "previous_heading"
        essential_modules["orca.cmdnames"].HEADING_NEXT = "next_heading"
        essential_modules["orca.cmdnames"].HEADING_LIST = "list_headings"
        essential_modules["orca.cmdnames"].HEADING_AT_LEVEL_PREV = "previous_heading_level_%d"
        essential_modules["orca.cmdnames"].HEADING_AT_LEVEL_NEXT = "next_heading_level_%d"
        essential_modules["orca.cmdnames"].HEADING_AT_LEVEL_LIST = "list_headings_level_%d"
        essential_modules["orca.cmdnames"].IFRAME_PREV = "previous_iframe"
        essential_modules["orca.cmdnames"].IFRAME_NEXT = "next_iframe"
        essential_modules["orca.cmdnames"].IFRAME_LIST = "list_iframes"
        essential_modules["orca.cmdnames"].IMAGE_PREV = "previous_image"
        essential_modules["orca.cmdnames"].IMAGE_NEXT = "next_image"
        essential_modules["orca.cmdnames"].IMAGE_LIST = "list_images"
        essential_modules["orca.cmdnames"].LANDMARK_PREV = "previous_landmark"
        essential_modules["orca.cmdnames"].LANDMARK_NEXT = "next_landmark"
        essential_modules["orca.cmdnames"].LANDMARK_LIST = "list_landmarks"
        essential_modules["orca.cmdnames"].LIST_PREV = "previous_list"
        essential_modules["orca.cmdnames"].LIST_NEXT = "next_list"
        essential_modules["orca.cmdnames"].LIST_LIST = "list_lists"
        essential_modules["orca.cmdnames"].LIST_ITEM_PREV = "previous_list_item"
        essential_modules["orca.cmdnames"].LIST_ITEM_NEXT = "next_list_item"
        essential_modules["orca.cmdnames"].LIST_ITEM_LIST = "list_list_items"
        essential_modules["orca.cmdnames"].LIVE_REGION_PREV = "previous_live_region"
        essential_modules["orca.cmdnames"].LIVE_REGION_NEXT = "next_live_region"
        essential_modules["orca.cmdnames"].LIVE_REGION_LAST = "last_live_region"
        essential_modules["orca.cmdnames"].PARAGRAPH_PREV = "previous_paragraph"
        essential_modules["orca.cmdnames"].PARAGRAPH_NEXT = "next_paragraph"
        essential_modules["orca.cmdnames"].PARAGRAPH_LIST = "list_paragraphs"
        essential_modules["orca.cmdnames"].RADIO_BUTTON_PREV = "previous_radio_button"
        essential_modules["orca.cmdnames"].RADIO_BUTTON_NEXT = "next_radio_button"
        essential_modules["orca.cmdnames"].RADIO_BUTTON_LIST = "list_radio_buttons"
        essential_modules["orca.cmdnames"].SEPARATOR_PREV = "previous_separator"
        essential_modules["orca.cmdnames"].SEPARATOR_NEXT = "next_separator"
        essential_modules["orca.cmdnames"].TABLE_PREV = "previous_table"
        essential_modules["orca.cmdnames"].TABLE_NEXT = "next_table"
        essential_modules["orca.cmdnames"].TABLE_LIST = "list_tables"
        essential_modules["orca.cmdnames"].LINK_PREV = "previous_link"
        essential_modules["orca.cmdnames"].LINK_NEXT = "next_link"
        essential_modules["orca.cmdnames"].LINK_LIST = "list_links"
        essential_modules["orca.cmdnames"].UNVISITED_LINK_PREV = "previous_unvisited_link"
        essential_modules["orca.cmdnames"].UNVISITED_LINK_NEXT = "next_unvisited_link"
        essential_modules["orca.cmdnames"].UNVISITED_LINK_LIST = "list_unvisited_links"
        essential_modules["orca.cmdnames"].VISITED_LINK_PREV = "previous_visited_link"
        essential_modules["orca.cmdnames"].VISITED_LINK_NEXT = "next_visited_link"
        essential_modules["orca.cmdnames"].VISITED_LINK_LIST = "list_visited_links"
        essential_modules["orca.cmdnames"].LARGE_OBJECT_PREV = "previous_large_object"
        essential_modules["orca.cmdnames"].LARGE_OBJECT_NEXT = "next_large_object"
        essential_modules["orca.cmdnames"].LARGE_OBJECT_LIST = "list_large_objects"
        essential_modules["orca.cmdnames"].CLICKABLE_PREV = "previous_clickable"
        essential_modules["orca.cmdnames"].CLICKABLE_NEXT = "next_clickable"
        essential_modules["orca.cmdnames"].CLICKABLE_LIST = "list_clickables"

    def test_init(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator initialization through get_navigator function."""

        self._setup_dependencies(test_context)
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        assert (
            nav._last_input_event is None or nav._last_input_event is not None
        )  # May be set by other tests
        assert isinstance(nav._suspended, bool)
        assert isinstance(nav._handlers, dict)
        assert nav._bindings is not None
        assert isinstance(nav._mode_for_script, dict)

    @pytest.mark.parametrize(
        "refresh_value, expects_setup_call, expects_debug_message",
        [
            pytest.param(True, True, True, id="refresh_true_calls_setup"),
            pytest.param(False, False, False, id="refresh_false_no_setup"),
            pytest.param(None, False, False, id="default_no_setup"),
        ],
    )
    def test_get_handlers_refresh_scenarios(
        self,
        test_context: OrcaTestContext,
        refresh_value,
        expects_setup_call,
        expects_debug_message,
    ) -> None:
        """Test StructuralNavigator.get_handlers with various refresh scenarios."""
        essential_modules = self._setup_dependencies(test_context)
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_setup = test_context.Mock()
        test_context.patch_object(nav, "_setup_handlers", new=mock_setup)

        if refresh_value is None:
            handlers = nav.get_handlers()
        else:
            handlers = nav.get_handlers(refresh=refresh_value)

        assert isinstance(handlers, dict)

        if expects_setup_call:
            mock_setup.assert_called_once()
        else:
            mock_setup.assert_not_called()

        if expects_debug_message:
            essential_modules["orca.debug"].print_message.assert_any_call(
                essential_modules["orca.debug"].LEVEL_INFO,
                "STRUCTURAL NAVIGATOR: Refreshing handlers.",
                True,
                True,
            )

    def test_setup_handlers(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator._setup_handlers creates handler dictionary."""

        self._setup_dependencies(test_context)
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        nav._setup_handlers()
        assert isinstance(nav._handlers, dict)
        assert len(nav._handlers) > 0
        assert "structural_navigator_mode_cycle" in nav._handlers
        assert "previous_button" in nav._handlers
        assert "next_button" in nav._handlers
        assert "list_buttons" in nav._handlers

    def test_setup_handlers_creates_navigation_handlers(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test StructuralNavigator._setup_handlers creates expected navigation handlers."""

        self._setup_dependencies(test_context)
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        nav._setup_handlers()
        expected_prefixes = ["previous_", "next_", "list_"]
        expected_elements = [
            "blockquote",
            "button",
            "checkbox",
            "combobox",
            "entry",
            "form_field",
            "heading",
            "iframe",
            "image",
            "landmark",
            "list",
            "list_item",
            "live_region",
            "paragraph",
            "radio_button",
            "separator",
            "table",
            "link",
            "large_object",
            "clickable",
        ]
        for element in expected_elements:
            for prefix in expected_prefixes:
                if prefix == "list_" and element in ["separator"]:
                    continue
                handler_key = f"{prefix}{element}"
                if element == "list" and prefix == "list_":
                    handler_key = "list_lists"
                elif element == "list_item" and prefix == "list_":
                    handler_key = "list_list_items"
                if handler_key in nav._handlers:
                    assert nav._handlers[handler_key] is not None

    def test_setup_handlers_creates_heading_level_handlers(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test StructuralNavigator._setup_handlers creates heading level handlers."""

        self._setup_dependencies(test_context)
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        nav._setup_handlers()
        for level in range(1, 7):
            assert f"previous_heading_level_{level}" in nav._handlers
            assert f"next_heading_level_{level}" in nav._handlers
            assert f"list_headings_level_{level}" in nav._handlers

    @pytest.mark.parametrize(
        "current_mode, expected_next_mode, supports_collection",
        [
            pytest.param("OFF", "DOCUMENT", True, id="off_to_document"),
            pytest.param("DOCUMENT", "GUI", True, id="document_to_gui"),
            pytest.param("GUI", "OFF", False, id="gui_to_off"),
        ],
    )
    def test_cycle_mode_transitions(
        self, test_context: OrcaTestContext, current_mode, expected_next_mode, supports_collection
    ) -> None:
        """Test StructuralNavigator.cycle_mode transitions."""

        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.structural_navigator import get_navigator, NavigationMode

        nav = get_navigator()
        current_nav_mode = getattr(NavigationMode, current_mode)
        expected_nav_mode = getattr(NavigationMode, expected_next_mode)

        mock_get_mode, mock_set_mode = self._setup_cycle_navigation_mode_mocks(
            test_context, nav, current_nav_mode, supports_collection=supports_collection
        )[:2]

        if supports_collection:
            essential_modules["orca.AXObject"].supports_collection.return_value = True

        result = nav.cycle_mode(mock_script, None, True)
        assert result is True
        mock_get_mode.assert_called_once_with(mock_script)
        mock_set_mode.assert_called_once_with(mock_script, expected_nav_mode)
        mock_script.present_message.assert_called()

    @pytest.mark.parametrize(
        "scenario, notify_user, is_active_script, expected_result, expects_message_call",
        [
            pytest.param("no_notify", False, True, True, False, id="no_notify_no_message"),
            pytest.param(
                "inactive_script", True, False, False, None, id="inactive_script_returns_false"
            ),
        ],
    )
    def test_cycle_mode_edge_scenarios(
        self,
        test_context: OrcaTestContext,
        scenario,
        notify_user,
        is_active_script,
        expected_result,
        expects_message_call,
    ) -> None:
        """Test StructuralNavigator.cycle_mode edge scenarios."""

        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.structural_navigator import get_navigator, NavigationMode

        nav = get_navigator()

        if scenario == "no_notify":
            self._setup_cycle_navigation_mode_mocks(test_context, nav, NavigationMode.OFF)
            essential_modules["orca.AXObject"].supports_collection.return_value = True
        else:
            mock_is_active_script = test_context.Mock(return_value=is_active_script)
            test_context.patch_object(nav, "_is_active_script", new=mock_is_active_script)

        result = nav.cycle_mode(mock_script, None, notify_user)
        assert result is expected_result

        if expects_message_call is False:
            mock_script.present_message.assert_not_called()
        elif scenario == "inactive_script":
            mock_is_active_script.assert_called_once_with(mock_script)

    @pytest.mark.parametrize(
        "suspend_value, is_active, initial_suspended, expected_suspended, expects_debug, "
        "expects_refresh",
        [
            pytest.param(True, True, False, True, True, True, id="suspend_navigation"),
            pytest.param(False, True, True, False, True, True, id="resume_navigation"),
            pytest.param(
                True,
                False,
                False,
                False,
                False,
                False,
                id="inactive_script_no_change",
            ),
        ],
    )
    def test_suspend_commands_scenarios(
        self,
        test_context: OrcaTestContext,
        suspend_value,
        is_active,
        initial_suspended,
        expected_suspended,
        expects_debug,
        expects_refresh,
    ) -> None:
        """Test StructuralNavigator.suspend_commands various scenarios."""

        essential_modules = (
            self._setup_dependencies(test_context)
            if expects_debug
            else self._setup_dependencies(test_context)
        )
        mock_script = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        nav._suspended = initial_suspended
        test_context.patch_object(nav, "_is_active_script", return_value=is_active)
        mock_refresh = test_context.Mock()
        test_context.patch_object(nav, "refresh_bindings_and_grabs", new=mock_refresh)

        nav.suspend_commands(mock_script, suspend_value, "test reason")

        assert nav._suspended == expected_suspended

        if expects_debug:
            essential_modules["orca.debug"].print_message.assert_any_call(
                essential_modules["orca.debug"].LEVEL_INFO,
                f"STRUCTURAL NAVIGATOR: Suspended: {suspend_value}: test reason",
                True,
            )

        if expects_refresh:
            mock_refresh.assert_called_once_with(
                mock_script, f"Suspended changed to {suspend_value}"
            )
        else:
            mock_refresh.assert_not_called()

    def test_get_object_in_direction_empty_objects(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator._get_object_in_direction with empty objects list."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        result = nav._get_object_in_direction(mock_script, [], True)
        assert result is None

    def test_get_object_in_direction_next_object(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator._get_object_in_direction returns next object."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_obj1 = test_context.Mock()
        mock_obj2 = test_context.Mock()
        mock_obj3 = test_context.Mock()
        objects = [mock_obj1, mock_obj2, mock_obj3]

        essential_modules[
            "orca.focus_manager"
        ].get_manager.return_value.get_locus_of_focus.return_value = mock_obj2
        essential_modules["orca.AXObject"].get_parent.return_value = mock_obj2

        def mock_get_path(obj):
            if obj == mock_obj1:
                return [0, 0]
            if obj == mock_obj2:
                return [0, 1]
            if obj == mock_obj3:
                return [0, 2]
            return []

        essential_modules["orca.AXObject"].get_path.side_effect = mock_get_path

        mock_script = test_context.Mock()

        def mock_path_comparison(path1, _path2):
            if path1 == [0, 0]:  # obj1
                return -1
            if path1 == [0, 1]:  # obj2
                return 0
            if path1 == [0, 2]:  # obj3
                return 1
            return 0

        mock_script.utilities.path_comparison.side_effect = mock_path_comparison

        result = nav._get_object_in_direction(mock_script, objects, True)
        assert result == mock_obj3

    def test_get_object_in_direction_previous_object(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator._get_object_in_direction returns previous object."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_obj1 = test_context.Mock()
        mock_obj2 = test_context.Mock()
        mock_obj3 = test_context.Mock()
        objects = [mock_obj1, mock_obj2, mock_obj3]

        essential_modules[
            "orca.focus_manager"
        ].get_manager.return_value.get_locus_of_focus.return_value = mock_obj2
        essential_modules["orca.AXObject"].get_parent.return_value = mock_obj2

        def mock_get_path(obj):
            if obj == mock_obj1:
                return [0, 0]
            if obj == mock_obj2:
                return [0, 1]
            if obj == mock_obj3:
                return [0, 2]  # Later in tree
            return []

        essential_modules["orca.AXObject"].get_path.side_effect = mock_get_path

        mock_script = test_context.Mock()

        def mock_path_comparison(path1, _path2):
            if path1 == [0, 0]:  # obj1
                return -1
            if path1 == [0, 1]:  # obj2
                return 0
            if path1 == [0, 2]:  # obj3
                return 1
            return 0

        mock_script.utilities.path_comparison.side_effect = mock_path_comparison

        result = nav._get_object_in_direction(mock_script, objects, False)
        assert result == mock_obj1

    def test_get_object_in_direction_wrap_to_beginning(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator._get_object_in_direction wraps to beginning when at end."""

        essential_modules = self._setup_dependencies(test_context)
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.return_value = True
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_result = test_context.Mock()
        mock_script = test_context.Mock()

        test_context.patch_object(nav, "_get_object_in_direction", return_value=mock_result)

        result = nav._get_object_in_direction(mock_script, [], True, True)
        assert result == mock_result

    def test_get_object_in_direction_wrap_to_end(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator._get_object_in_direction wraps to end when at beginning."""

        essential_modules = self._setup_dependencies(test_context)
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.return_value = True
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_result = test_context.Mock()
        mock_script = test_context.Mock()

        test_context.patch_object(nav, "_get_object_in_direction", return_value=mock_result)

        result = nav._get_object_in_direction(mock_script, [], False, True)
        assert result == mock_result

    def test_get_object_in_direction_no_wrap_at_end(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator._get_object_in_direction returns None at end when no wrap."""

        self._setup_dependencies(test_context)
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_script = test_context.Mock()

        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)

        result = nav._get_object_in_direction(mock_script, [], True, False)
        assert result is None

    def test_previous_button_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.previous_button method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_get_buttons = test_context.Mock(return_value=[])
        test_context.patch_object(nav, "_get_all_buttons", new=mock_get_buttons)
        mock_get_in_direction = test_context.Mock(return_value=None)
        test_context.patch_object(nav, "_get_object_in_direction", new=mock_get_in_direction)
        mock_present = test_context.Mock()
        test_context.patch_object(nav, "_present_object", new=mock_present)
        result = nav.previous_button(mock_script, mock_event, True)
        assert result is True
        mock_get_buttons.assert_called_once_with(mock_script)
        mock_get_in_direction.assert_called_once()
        mock_present.assert_called_once()

    def test_next_button_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.next_button method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_get_buttons = test_context.Mock(return_value=[])
        test_context.patch_object(nav, "_get_all_buttons", new=mock_get_buttons)
        mock_get_in_direction = test_context.Mock(return_value=None)
        test_context.patch_object(nav, "_get_object_in_direction", new=mock_get_in_direction)
        mock_present = test_context.Mock()
        test_context.patch_object(nav, "_present_object", new=mock_present)
        result = nav.next_button(mock_script, mock_event, True)
        assert result is True
        mock_get_buttons.assert_called_once_with(mock_script)
        mock_get_in_direction.assert_called_once()
        mock_present.assert_called_once()

    def test_list_buttons_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.list_buttons method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_get_buttons = test_context.Mock(return_value=[])
        test_context.patch_object(nav, "_get_all_buttons", new=mock_get_buttons)
        mock_present = test_context.Mock()
        test_context.patch_object(nav, "_present_object_list", new=mock_present)
        result = nav.list_buttons(mock_script, mock_event)
        assert result is True
        mock_get_buttons.assert_called_once_with(mock_script)
        mock_present.assert_called_once()

    def test_get_bindings_refresh_true(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.get_bindings with refresh=True updates bindings."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_setup = test_context.Mock()
        test_context.patch_object(nav, "_setup_bindings", new=mock_setup)
        bindings = nav.get_bindings(refresh=True)
        assert bindings.key_bindings is not None
        essential_modules["orca.debug"].print_message.assert_any_call(
            essential_modules["orca.debug"].LEVEL_INFO,
            "STRUCTURAL NAVIGATOR: Refreshing bindings. Is desktop: True",
            True,
        )
        mock_setup.assert_called_once()

    def test_get_bindings_refresh_false(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.get_bindings with refresh=False returns cached bindings."""

        self._setup_dependencies(test_context)
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_bindings = test_context.Mock()
        mock_bindings.key_bindings = []
        test_context.patch_object(nav, "get_bindings", return_value=mock_bindings)

        bindings = nav.get_bindings(refresh=False)
        assert bindings.key_bindings is not None
        assert bindings == mock_bindings

    def test_setup_bindings(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator._setup_bindings creates keybindings."""
        self._setup_dependencies(test_context)
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_bindings = test_context.Mock()
        mock_bindings.key_bindings = []

        def mock_setup_bindings():
            nav._bindings = mock_bindings

        test_context.patch_object(nav, "_setup_bindings", new=mock_setup_bindings)
        nav._setup_bindings()
        assert nav._bindings.key_bindings is not None
        assert nav._bindings == mock_bindings

    @pytest.mark.parametrize(
        "is_same_script,expected",
        [
            (True, True),
            (False, False),
        ],
    )
    def test_is_active_script(
        self, test_context: OrcaTestContext, is_same_script: bool, expected: bool
    ) -> None:
        """Test StructuralNavigator._is_active_script with active and inactive scripts."""

        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_active_script = mock_script if is_same_script else test_context.Mock()
        essential_modules[
            "orca.script_manager"
        ].get_manager.return_value.get_active_script.return_value = mock_active_script
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        result = nav._is_active_script(mock_script)
        assert result is expected

    def test_get_mode_default(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.get_mode returns default mode for new script."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.structural_navigator import get_navigator, NavigationMode

        nav = get_navigator()
        result = nav.get_mode(mock_script)
        assert result == NavigationMode.OFF

    def test_get_mode_cached(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.get_mode returns cached mode for known script."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.structural_navigator import get_navigator, NavigationMode

        nav = get_navigator()
        nav._mode_for_script[mock_script] = NavigationMode.DOCUMENT
        result = nav.get_mode(mock_script)
        assert result == NavigationMode.DOCUMENT

    def test_set_mode(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.set_mode sets mode for script."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.structural_navigator import get_navigator, NavigationMode

        nav = get_navigator()
        nav.set_mode(mock_script, NavigationMode.GUI)
        assert nav._mode_for_script[mock_script] == NavigationMode.GUI

    def test_determine_root_container_document_mode(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator._determine_root_container for document mode."""
        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_focus = test_context.Mock()
        mock_document = test_context.Mock()

        essential_modules[
            "orca.focus_manager"
        ].get_manager().get_locus_of_focus.return_value = mock_focus
        essential_modules["orca.AXObject"].find_ancestor_inclusive.side_effect = (
            lambda focus, predicate: None
        )
        mock_script.utilities.get_top_level_document_for_object.return_value = mock_document

        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_determine_root_container", return_value=mock_document)

        result = nav._determine_root_container(mock_script)
        assert result == mock_document

    def test_determine_root_container_fallback(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator._determine_root_container fallback to app."""

        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_app = test_context.Mock(spec=Atspi.Accessible)
        mock_script.app = mock_app
        essential_modules["orca.AXUtilities"].is_web_application.return_value = False
        essential_modules["orca.AXObject"].find_ancestor_inclusive.return_value = mock_app
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_determine_root_container", return_value=mock_app)
        result = nav._determine_root_container(mock_script)
        assert result == mock_app

    @pytest.mark.parametrize(
        "element_type,ax_method",
        [
            pytest.param("buttons", "find_all_buttons", id="get_all_buttons"),
            pytest.param("headings", "find_all_headings", id="get_all_headings"),
            pytest.param("links", "find_all_links", id="get_all_links"),
            pytest.param("tables", "find_all_tables", id="get_all_tables"),
            pytest.param("lists", "find_all_lists", id="get_all_lists"),
            pytest.param("checkboxes", "find_all_check_boxes", id="get_all_checkboxes"),
            pytest.param("entries", "find_all_editable_objects", id="get_all_entries"),
            pytest.param("form_fields", "find_all_form_fields", id="get_all_form_fields"),
            pytest.param("images", "find_all_images_and_image_maps", id="get_all_images"),
            pytest.param("landmarks", "find_all_landmarks", id="get_all_landmarks"),
            pytest.param("paragraphs", "find_all_paragraphs", id="get_all_paragraphs"),
        ],
    )
    def test_get_all_elements(
        self,
        test_context: OrcaTestContext,
        element_type: str,
        ax_method: str,
    ) -> None:
        """Test StructuralNavigator._get_all_* methods with various element types."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_elements = [test_context.Mock(), test_context.Mock()]

        if element_type == "lists":
            test_context.patch(
                f"orca.structural_navigator.AXUtilities.{ax_method}",
                side_effect=lambda root,
                include_description_lists=False,
                include_tab_lists=False,
                pred=None: mock_elements,
            )
        elif element_type == "paragraphs":
            test_context.patch(
                f"orca.structural_navigator.AXUtilities.{ax_method}",
                side_effect=lambda root, include_items=False, pred=None: mock_elements,
            )
        else:
            test_context.patch(
                f"orca.structural_navigator.AXUtilities.{ax_method}",
                side_effect=lambda root, pred=None: mock_elements,
            )

        if element_type in ["links", "entries", "images"]:
            mock_root = test_context.Mock()
            test_context.patch_object(
                nav, "_determine_root_container", side_effect=lambda script: mock_root
            )

        navigation_method = getattr(nav, f"_get_all_{element_type}")
        result = navigation_method(mock_script)
        assert result == mock_elements

    def test_get_mode(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.get_mode method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.structural_navigator import get_navigator, NavigationMode

        nav = get_navigator()
        result = nav.get_mode(mock_script)
        assert result == NavigationMode.OFF

    def test_last_input_event_was_navigation_command(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.last_input_event_was_navigation_command method."""

        essential_modules = self._setup_dependencies(test_context)
        essential_modules[
            "orca.input_event_manager"
        ].get_manager().last_event_equals_or_is_release_for_event.return_value = True
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_event = test_context.Mock()
        mock_event.as_single_line_string = test_context.Mock(return_value="mock event")
        nav._last_input_event = mock_event

        result = nav.last_input_event_was_navigation_command()
        assert isinstance(result, bool)
        assert result is True

    def test_get_state_string(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator._get_state_string method."""

        essential_modules = self._setup_dependencies(test_context)
        mock_obj = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        essential_modules["orca.AXObject"].get_state_set.return_value = None
        essential_modules["orca.AXUtilities"].is_switch.return_value = False
        essential_modules["orca.AXUtilities"].is_checked.return_value = False
        essential_modules["orca.AXUtilities"].is_expanded.return_value = False
        essential_modules["orca.AXUtilities"].is_pressed.return_value = False
        essential_modules["orca.AXUtilities"].is_selected.return_value = False
        test_context.patch_object(nav, "_get_state_string", return_value="test state")
        result = nav._get_state_string(mock_obj)
        assert isinstance(result, str)
        assert result == "test state"

    def test_get_item_string(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator._get_item_string method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_obj = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(
            nav, "_get_item_string", return_value="Test Object: Test Description"
        )
        result = nav._get_item_string(mock_script, mock_obj)
        assert isinstance(result, str)
        assert "Test Description" in result

    def test_is_non_document_object(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator._is_non_document_object method."""

        essential_modules = self._setup_dependencies(test_context)
        mock_obj = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        essential_modules["orca.AXUtilities"].is_application.return_value = False
        essential_modules["orca.AXUtilities"].is_document.return_value = False
        essential_modules["orca.AXUtilities"].is_frame.return_value = False
        essential_modules["orca.AXObject"].get_role.return_value = None
        result = nav._is_non_document_object(mock_obj)
        assert isinstance(result, bool)

    def test_add_bindings(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.add_bindings method."""

        essential_modules = self._setup_dependencies(test_context)
        essential_modules["orca.debug"].debugLevel = 0
        mock_script = test_context.Mock()
        mock_key_bindings = test_context.Mock()
        mock_key_bindings.get_bindings_with_grabs_for_debugging.return_value = []
        mock_script.get_key_bindings.return_value = mock_key_bindings
        mock_script.key_bindings = mock_key_bindings
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_bindings = test_context.Mock()
        mock_bindings.key_bindings = []
        nav._bindings = mock_bindings
        test_context.patch_object(nav, "_is_active_script", return_value=True)
        test_context.patch_object(nav, "get_bindings", return_value=mock_bindings)
        nav.add_bindings(mock_script, "test")
        essential_modules["orca.debug"].print_tokens.assert_called()

    def test_remove_bindings(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.remove_bindings method."""

        essential_modules = self._setup_dependencies(test_context)
        essential_modules["orca.debug"].debugLevel = 0
        mock_script = test_context.Mock()
        mock_key_bindings = test_context.Mock()
        mock_key_bindings.get_bindings_with_grabs_for_debugging.return_value = []
        mock_script.get_key_bindings.return_value = mock_key_bindings
        mock_script.key_bindings = mock_key_bindings
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_bindings = test_context.Mock()
        mock_bindings.key_bindings = []
        nav._bindings = mock_bindings
        test_context.patch_object(nav, "_is_active_script", return_value=True)
        nav.remove_bindings(mock_script, "test")
        essential_modules["orca.debug"].print_tokens.assert_called()

    def test_refresh_bindings_and_grabs(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.refresh_bindings_and_grabs method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_is_active_script", return_value=True)
        test_context.patch_object(nav, "remove_bindings", new=test_context.Mock())
        test_context.patch_object(nav, "add_bindings", new=test_context.Mock())
        nav.refresh_bindings_and_grabs(mock_script, "test")

    def test_previous_heading_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.previous_heading method."""
        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_headings", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.previous_heading(mock_script, mock_event, True)
        assert result is True

    def test_next_heading_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.next_heading method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_headings", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.next_heading(mock_script, mock_event, True)
        assert result is True

    def test_list_headings_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.list_headings method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_headings", return_value=[])
        test_context.patch_object(nav, "_present_object_list", new=test_context.Mock())
        result = nav.list_headings(mock_script, mock_event)
        assert result is True

    def test_previous_link_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.previous_link method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_links", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.previous_link(mock_script, mock_event, True)
        assert result is True

    def test_next_link_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.next_link method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_links", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.next_link(mock_script, mock_event, True)
        assert result is True

    def test_list_links_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.list_links method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_links", return_value=[])
        test_context.patch_object(nav, "_present_object_list", new=test_context.Mock())
        result = nav.list_links(mock_script, mock_event)
        assert result is True

    def test_previous_table_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.previous_table method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_tables", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.previous_table(mock_script, mock_event, True)
        assert result is True

    def test_next_table_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.next_table method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_tables", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.next_table(mock_script, mock_event, True)
        assert result is True

    def test_list_tables_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.list_tables method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_tables", return_value=[])
        test_context.patch_object(nav, "_present_object_list", new=test_context.Mock())
        result = nav.list_tables(mock_script, mock_event)
        assert result is True

    def test_previous_list_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.previous_list method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_lists", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.previous_list(mock_script, mock_event, True)
        assert result is True

    def test_next_list_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.next_list method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_lists", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.next_list(mock_script, mock_event, True)
        assert result is True

    def test_list_lists_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.list_lists method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_lists", return_value=[])
        test_context.patch_object(nav, "_present_object_list", new=test_context.Mock())
        result = nav.list_lists(mock_script, mock_event)
        assert result is True

    def test_previous_blockquote_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.previous_blockquote method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_blockquotes", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.previous_blockquote(mock_script, mock_event, True)
        assert result is True

    def test_next_blockquote_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.next_blockquote method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_blockquotes", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.next_blockquote(mock_script, mock_event, True)
        assert result is True

    def test_previous_checkbox_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.previous_checkbox method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_checkboxes", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.previous_checkbox(mock_script, mock_event, True)
        assert result is True

    def test_next_checkbox_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.next_checkbox method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_checkboxes", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.next_checkbox(mock_script, mock_event, True)
        assert result is True

    def test_previous_entry_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.previous_entry method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_entries", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.previous_entry(mock_script, mock_event, True)
        assert result is True

    def test_next_entry_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.next_entry method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_entries", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.next_entry(mock_script, mock_event, True)
        assert result is True

    def test_previous_form_field_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.previous_form_field method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_form_fields", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.previous_form_field(mock_script, mock_event, True)
        assert result is True

    def test_next_form_field_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.next_form_field method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_form_fields", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.next_form_field(mock_script, mock_event, True)
        assert result is True

    def test_previous_image_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.previous_image method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_images", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.previous_image(mock_script, mock_event, True)
        assert result is True

    def test_next_image_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.next_image method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_images", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.next_image(mock_script, mock_event, True)
        assert result is True

    def test_previous_landmark_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.previous_landmark method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_landmarks", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.previous_landmark(mock_script, mock_event, True)
        assert result is True

    def test_next_landmark_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.next_landmark method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_landmarks", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.next_landmark(mock_script, mock_event, True)
        assert result is True

    def test_previous_paragraph_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.previous_paragraph method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_paragraphs", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.previous_paragraph(mock_script, mock_event, True)
        assert result is True

    def test_next_paragraph_method(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.next_paragraph method."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_event = test_context.Mock()
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_get_all_paragraphs", return_value=[])
        test_context.patch_object(nav, "_get_object_in_direction", return_value=None)
        test_context.patch_object(nav, "_present_object", new=test_context.Mock())
        result = nav.next_paragraph(mock_script, mock_event, True)
        assert result is True

    @pytest.mark.parametrize(
        "expected_next_method,expected_prev_method",
        [
            pytest.param("next_heading_level_1", "previous_heading_level_1", id="heading_level_1"),
            pytest.param("next_heading_level_2", "previous_heading_level_2", id="heading_level_2"),
            pytest.param("next_heading_level_3", "previous_heading_level_3", id="heading_level_3"),
            pytest.param("next_heading_level_4", "previous_heading_level_4", id="heading_level_4"),
            pytest.param("next_heading_level_5", "previous_heading_level_5", id="heading_level_5"),
            pytest.param("next_heading_level_6", "previous_heading_level_6", id="heading_level_6"),
        ],
    )
    def test_heading_level_navigation_methods_exist(
        self,
        test_context: OrcaTestContext,
        expected_next_method: str,
        expected_prev_method: str,
    ) -> None:
        """Test StructuralNavigator has navigation methods for each heading level."""

        essential_modules = test_context.setup_shared_dependencies(
            [
                "orca.keybindings",
                "orca.settings_manager",
                "orca.dbus_service",
                "orca.debug",
                "orca.focus_manager",
                "orca.AXObject",
                "orca.AXUtilities",
            ]
        )
        self._setup_mocks(test_context, essential_modules)

        from orca.structural_navigator import StructuralNavigator

        navigator = StructuralNavigator()
        assert hasattr(navigator, expected_next_method)
        assert hasattr(navigator, expected_prev_method)
        assert callable(getattr(navigator, expected_next_method))
        assert callable(getattr(navigator, expected_prev_method))

    @pytest.mark.parametrize(
        "mode_value",
        [
            pytest.param("OFF", id="off_mode"),
            pytest.param("DOCUMENT", id="document_mode"),
            pytest.param("GUI", id="gui_mode"),
        ],
    )
    def test_navigation_mode_values(
        self,
        test_context: OrcaTestContext,
        mode_value: str,
    ) -> None:
        """Test NavigationMode enum values."""

        essential_modules = test_context.setup_shared_dependencies(
            [
                "orca.keybindings",
                "orca.settings_manager",
                "orca.dbus_service",
                "orca.debug",
                "orca.focus_manager",
                "orca.AXObject",
                "orca.AXUtilities",
            ]
        )
        self._setup_mocks(test_context, essential_modules)

        from orca.structural_navigator import NavigationMode

        if mode_value == "OFF":
            assert NavigationMode.OFF.value == "OFF"
        elif mode_value == "DOCUMENT":
            assert NavigationMode.DOCUMENT.value == "DOCUMENT"
        elif mode_value == "GUI":
            assert NavigationMode.GUI.value == "GUI"

    def test_basic_structural_navigator_state(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator basic state management."""

        essential_modules = test_context.setup_shared_dependencies(
            [
                "orca.keybindings",
                "orca.settings_manager",
                "orca.dbus_service",
                "orca.debug",
                "orca.focus_manager",
                "orca.AXObject",
                "orca.AXUtilities",
            ]
        )
        self._setup_mocks(test_context, essential_modules)

        from orca.structural_navigator import StructuralNavigator

        navigator = StructuralNavigator()
        assert navigator._last_input_event is None
        assert navigator._suspended is False
        assert isinstance(navigator._mode_for_script, dict)

    def test_get_is_enabled(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.get_is_enabled returns setting value."""

        essential_modules = self._setup_dependencies(test_context)
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.return_value = True
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        result = nav.get_is_enabled()
        assert result is True
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.assert_called_with("structuralNavigationEnabled")

    def test_set_is_enabled_no_change(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.set_is_enabled returns early if value unchanged."""

        essential_modules = self._setup_dependencies(test_context)
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.return_value = True
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        result = nav.set_is_enabled(True)
        assert result is True
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.set_setting.assert_not_called()

    def test_set_is_enabled_true_with_previous_mode(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.set_is_enabled restores previous mode when enabling."""

        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.return_value = False
        essential_modules[
            "orca.script_manager"
        ].get_manager.return_value.get_active_script.return_value = mock_script
        from orca.structural_navigator import get_navigator, NavigationMode

        nav = get_navigator()
        nav._previous_mode_for_script[mock_script] = NavigationMode.DOCUMENT
        test_context.patch_object(nav, "_is_active_script", return_value=True)
        mock_refresh = test_context.Mock()
        test_context.patch_object(nav, "refresh_bindings_and_grabs", new=mock_refresh)

        result = nav.set_is_enabled(True)
        assert result is True
        assert nav._mode_for_script[mock_script] == NavigationMode.DOCUMENT
        mock_refresh.assert_called_once()

    def test_set_is_enabled_true_without_previous_mode(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.set_is_enabled without previous mode when enabling."""

        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.return_value = False
        essential_modules[
            "orca.script_manager"
        ].get_manager.return_value.get_active_script.return_value = mock_script
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        test_context.patch_object(nav, "_is_active_script", return_value=True)
        mock_refresh = test_context.Mock()
        test_context.patch_object(nav, "refresh_bindings_and_grabs", new=mock_refresh)

        result = nav.set_is_enabled(True)
        assert result is True
        mock_refresh.assert_called_once()

    def test_set_is_enabled_false_saves_mode(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.set_is_enabled saves current mode when disabling."""

        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.return_value = True
        essential_modules[
            "orca.script_manager"
        ].get_manager.return_value.get_active_script.return_value = mock_script
        from orca.structural_navigator import get_navigator, NavigationMode

        nav = get_navigator()
        nav._mode_for_script[mock_script] = NavigationMode.DOCUMENT
        test_context.patch_object(nav, "_is_active_script", return_value=True)
        mock_refresh = test_context.Mock()
        test_context.patch_object(nav, "refresh_bindings_and_grabs", new=mock_refresh)

        result = nav.set_is_enabled(False)
        assert result is True
        assert nav._previous_mode_for_script[mock_script] == NavigationMode.DOCUMENT
        assert nav._mode_for_script[mock_script] == NavigationMode.OFF
        mock_refresh.assert_called_once()

    def test_set_is_enabled_false_already_off(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.set_is_enabled returns early if already OFF."""

        essential_modules = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.return_value = True
        essential_modules[
            "orca.script_manager"
        ].get_manager.return_value.get_active_script.return_value = mock_script
        from orca.structural_navigator import get_navigator, NavigationMode

        nav = get_navigator()
        nav._mode_for_script[mock_script] = NavigationMode.OFF
        mock_refresh = test_context.Mock()
        test_context.patch_object(nav, "refresh_bindings_and_grabs", new=mock_refresh)

        result = nav.set_is_enabled(False)
        assert result is True
        mock_refresh.assert_not_called()

    def test_get_triggers_focus_mode(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.get_triggers_focus_mode returns setting value."""

        essential_modules = self._setup_dependencies(test_context)
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.return_value = False
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        result = nav.get_triggers_focus_mode()
        assert result is False
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.assert_called_with("structNavTriggersFocusMode")

    def test_set_triggers_focus_mode(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.set_triggers_focus_mode updates setting."""

        essential_modules = self._setup_dependencies(test_context)
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.return_value = True
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        result = nav.set_triggers_focus_mode(False)
        assert result is True
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.set_setting.assert_called_with(
            "structNavTriggersFocusMode", False
        )

    def test_set_triggers_focus_mode_no_change(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.set_triggers_focus_mode returns early if unchanged."""

        essential_modules = self._setup_dependencies(test_context)
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.return_value = True
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        result = nav.set_triggers_focus_mode(True)
        assert result is True
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.set_setting.assert_not_called()

    def test_last_command_prevents_focus_mode_true(self, test_context: OrcaTestContext) -> None:
        """Test StructuralNavigator.last_command_prevents_focus_mode returns True."""

        essential_modules = self._setup_dependencies(test_context)
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.return_value = False
        essential_modules[
            "orca.input_event_manager"
        ].get_manager.return_value.last_event_equals_or_is_release_for_event.return_value = True
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_event = test_context.Mock()
        nav._last_input_event = mock_event
        result = nav.last_command_prevents_focus_mode()
        assert result is True

    def test_last_command_prevents_focus_mode_false_no_event(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test StructuralNavigator.last_command_prevents_focus_mode returns False if no event."""

        self._setup_dependencies(test_context)
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        nav._last_input_event = None
        result = nav.last_command_prevents_focus_mode()
        assert result is False

    def test_last_command_prevents_focus_mode_false_setting_true(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test StructuralNavigator.last_command_prevents_focus_mode returns False if setting True."""

        essential_modules = self._setup_dependencies(test_context)
        essential_modules[
            "orca.settings_manager"
        ].get_manager.return_value.get_setting.return_value = True
        from orca.structural_navigator import get_navigator

        nav = get_navigator()
        mock_event = test_context.Mock()
        nav._last_input_event = mock_event
        result = nav.last_command_prevents_focus_mode()
        assert result is False



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2016-2023 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-lines
# pylint: disable=wrong-import-position
# pylint: disable=c-extension-no-member
# pylint: disable=too-many-public-methods

"""Module for flat-review commands"""

# This has to be the first non-docstring line in the module to make linters happy.
from __future__ import annotations

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2005-2008 Sun Microsystems Inc." \
                "Copyright (c) 2016-2023 Igalia, S.L."
__license__   = "LGPL"

import time
from typing import TYPE_CHECKING

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from . import braille
from . import cmdnames
from . import dbus_service
from . import debug
from . import flat_review
from . import focus_manager
from . import guilabels
from . import input_event
from . import keybindings
from . import messages
from . import script_manager
from . import settings_manager
from . import settings
from . import speech_and_verbosity_manager
from .ax_event_synthesizer import AXEventSynthesizer
from .ax_object import AXObject
from .ax_text import AXText

if TYPE_CHECKING:
    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from .scripts import default


class FlatReviewPresenter:
    """Provides access to on-screen objects via flat-review."""

    def __init__(self) -> None:
        self._context: flat_review.Context | None = None
        self._current_contents: str = ""
        self._restrict: bool = \
            settings_manager.get_manager().get_setting("flatReviewIsRestricted") or False
        self._handlers: dict[str, input_event.InputEventHandler] = self.get_handlers(True)
        self._desktop_bindings: keybindings.KeyBindings = keybindings.KeyBindings()
        self._laptop_bindings: keybindings.KeyBindings = keybindings.KeyBindings()
        self._gui: FlatReviewContextGUI | None = None

        msg = "FLAT REVIEW PRESENTER: Registering D-Bus commands."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        controller = dbus_service.get_remote_controller()
        controller.register_decorated_module("FlatReviewPresenter", self)

    def is_active(self) -> bool:
        """Returns True if the flat review presenter is active."""

        return self._context is not None

    def get_or_create_context(self, script: default.Script | None = None) -> flat_review.Context:
        """Returns the flat review context, creating one if necessary."""

        if not self._context:
            msg = f"FLAT REVIEW PRESENTER: Creating new context. Restrict: {self._restrict}"
            debug.print_message(debug.LEVEL_INFO, msg, True)

            if self._restrict:
                mode, obj = focus_manager.get_manager().get_active_mode_and_object_of_interest()
                self._context = flat_review.Context(script, root=obj)
            else:
                self._context = flat_review.Context(script)

            focus_manager.get_manager().emit_region_changed(
                self._context.get_current_object(), mode=focus_manager.FLAT_REVIEW)

            return self._context

        msg = f"FLAT REVIEW PRESENTER: Using existing context. Restrict: {self._restrict}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        # If we are in unrestricted mode, update the context as below.
        # If the context already exists, but the active mode is not flat review, update
        # the flat review location to that of the object of interest -- if the object of
        # interest is in the flat review context (which means it's on screen). In some
        # cases the object of interest will not be in the flat review context because it
        # is represented by descendant text objects. set_current_to_zone_with_object checks
        # for this condition and if it can find a zone whose ancestor is the object of
        # interest, it will set the current zone to the descendant, causing Orca to
        # present the text at the location of the object of interest.
        mode, obj = focus_manager.get_manager().get_active_mode_and_object_of_interest()
        obj = obj or focus_manager.get_manager().get_locus_of_focus()
        if mode != focus_manager.FLAT_REVIEW and obj != self._context.get_current_object() \
           and not self._restrict:
            tokens = ["FLAT REVIEW PRESENTER: Attempting to update location from",
                      self._context.get_current_object(), "to", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            self._context.set_current_to_zone_with_object(obj)

        # If we are restricting, and the current mode is not flat review, calculate a new context
        if self._restrict and mode != focus_manager.FLAT_REVIEW:
            msg = "FLAT REVIEW PRESENTER: Creating new restricted context."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._context = flat_review.Context(script, obj)

        return self._context

    def get_bindings(
        self,
        refresh: bool = False,
        is_desktop: bool = True
    ) -> keybindings.KeyBindings:
        """Returns the flat-review-presenter keybindings."""

        if refresh:
            msg = "FLAT REVIEW PRESENTER: Refreshing bindings."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._laptop_bindings.remove_key_grabs("FLAT REVIEW PRESENTER: Refreshing bindings.")
            self._desktop_bindings.remove_key_grabs("FLAT REVIEW PRESENTER: Refreshing bindings.")
            self._setup_bindings()
        elif is_desktop and self._desktop_bindings.is_empty():
            self._setup_bindings()
        elif not is_desktop and self._laptop_bindings.is_empty():
            self._setup_bindings()

        if is_desktop:
            return self._desktop_bindings
        return self._laptop_bindings

    def get_braille_bindings(self) -> dict[int, input_event.InputEventHandler]:
        """Returns the flat-review-presenter braille bindings."""

        bindings = {}
        try:
            bindings[braille.brlapi.KEY_CMD_LNUP] = \
                self._handlers["reviewAboveHandler"]
            bindings[braille.brlapi.KEY_CMD_LNDN] = \
                self._handlers["reviewBelowHandler"]
            bindings[braille.brlapi.KEY_CMD_FREEZE] = \
                self._handlers["toggleFlatReviewModeHandler"]
            bindings[braille.brlapi.KEY_CMD_TOP_LEFT] = \
                self._handlers["reviewHomeHandler"]
            bindings[braille.brlapi.KEY_CMD_BOT_LEFT] = \
                self._handlers["reviewBottomLeftHandler"]
        except AttributeError as error:
            tokens = ["FLAT REVIEW PRESENTER: Exception getting braille bindings:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return {}
        return bindings

    def get_handlers(self, refresh: bool = False) -> dict[str, input_event.InputEventHandler]:
        """Returns the flat-review-presenter handlers."""

        if refresh:
            msg = "FLAT REVIEW PRESENTER: Refreshing handlers."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._setup_handlers()

        return self._handlers

    def _setup_bindings(self) -> None:
        """Sets up the flat-review-presenter key bindings."""

        self._setup_desktop_bindings()
        self._setup_laptop_bindings()

    def _setup_handlers(self) -> None:
        """Sets up the flat-review-presenter input event handlers."""

        self._handlers = {}

        self._handlers["toggleFlatReviewModeHandler"] = \
            input_event.InputEventHandler(
                self.toggle_flat_review_mode,
                cmdnames.TOGGLE_FLAT_REVIEW)

        self._handlers["reviewHomeHandler"] = \
            input_event.InputEventHandler(
                self.go_home,
                cmdnames.REVIEW_HOME)

        self._handlers["reviewEndHandler"] = \
            input_event.InputEventHandler(
                self.go_end,
                cmdnames.REVIEW_END)

        self._handlers["reviewBottomLeftHandler"] = \
            input_event.InputEventHandler(
                self.go_bottom_left,
                cmdnames.REVIEW_BOTTOM_LEFT)

        self._handlers["reviewPreviousLineHandler"] = \
            input_event.InputEventHandler(
                self.go_previous_line,
                cmdnames.REVIEW_PREVIOUS_LINE)

        self._handlers["reviewCurrentLineHandler"] = \
            input_event.InputEventHandler(
                self.present_line,
                cmdnames.REVIEW_CURRENT_LINE)

        self._handlers["reviewNextLineHandler"] = \
            input_event.InputEventHandler(
                self.go_next_line,
                cmdnames.REVIEW_NEXT_LINE)

        self._handlers["reviewSpellCurrentLineHandler"] = \
            input_event.InputEventHandler(
                self.spell_line,
                cmdnames.REVIEW_SPELL_CURRENT_LINE)

        self._handlers["reviewPhoneticCurrentLineHandler"] = \
            input_event.InputEventHandler(
                self.phonetic_line,
                cmdnames.REVIEW_PHONETIC_CURRENT_LINE)

        self._handlers["reviewEndOfLineHandler"] = \
            input_event.InputEventHandler(
                self.go_end_of_line,
                cmdnames.REVIEW_END_OF_LINE)

        self._handlers["reviewPreviousItemHandler"] = \
            input_event.InputEventHandler(
                self.go_previous_item,
                cmdnames.REVIEW_PREVIOUS_ITEM)

        self._handlers["reviewCurrentItemHandler"] = \
            input_event.InputEventHandler(
                self.present_item,
                cmdnames.REVIEW_CURRENT_ITEM)

        self._handlers["reviewNextItemHandler"] = \
            input_event.InputEventHandler(
                self.go_next_item,
                cmdnames.REVIEW_NEXT_ITEM)

        self._handlers["reviewSpellCurrentItemHandler"] = \
            input_event.InputEventHandler(
                self.spell_item,
                cmdnames.REVIEW_SPELL_CURRENT_ITEM)

        self._handlers["reviewPhoneticCurrentItemHandler"] = \
            input_event.InputEventHandler(
                self.phonetic_item,
                cmdnames.REVIEW_PHONETIC_CURRENT_ITEM)

        self._handlers["reviewPreviousCharacterHandler"] = \
            input_event.InputEventHandler(
                self.go_previous_character,
                cmdnames.REVIEW_PREVIOUS_CHARACTER)

        self._handlers["reviewCurrentCharacterHandler"] = \
            input_event.InputEventHandler(
                self.present_character,
                cmdnames.REVIEW_CURRENT_CHARACTER)

        self._handlers["reviewSpellCurrentCharacterHandler"] = \
            input_event.InputEventHandler(
                self.spell_character,
                cmdnames.REVIEW_SPELL_CURRENT_CHARACTER)

        self._handlers["reviewUnicodeCurrentCharacterHandler"] = \
            input_event.InputEventHandler(
                self.unicode_current_character,
                cmdnames.REVIEW_UNICODE_CURRENT_CHARACTER)

        self._handlers["reviewNextCharacterHandler"] = \
            input_event.InputEventHandler(
                self.go_next_character,
                cmdnames.REVIEW_NEXT_CHARACTER)

        self._handlers["reviewCurrentAccessibleHandler"] = \
            input_event.InputEventHandler(
                self.present_object,
                cmdnames.REVIEW_CURRENT_ACCESSIBLE)

        self._handlers["reviewAboveHandler"] = \
            input_event.InputEventHandler(
                self.go_above,
                cmdnames.REVIEW_ABOVE)

        self._handlers["reviewBelowHandler"] = \
            input_event.InputEventHandler(
                self.go_below,
                cmdnames.REVIEW_BELOW)

        self._handlers["showContentsHandler"] = \
            input_event.InputEventHandler(
                self.show_contents,
                cmdnames.FLAT_REVIEW_SHOW_CONTENTS)

        self._handlers["flatReviewCopyHandler"] = \
            input_event.InputEventHandler(
                self.copy_to_clipboard,
                cmdnames.FLAT_REVIEW_COPY)

        self._handlers["flatReviewAppendHandler"] = \
            input_event.InputEventHandler(
                self.append_to_clipboard,
                cmdnames.FLAT_REVIEW_APPEND)

        self._handlers["flatReviewSayAllHandler"] = \
            input_event.InputEventHandler(
                self.say_all,
                cmdnames.SAY_ALL_FLAT_REVIEW)

        self._handlers["flatReviewToggleRestrictHandler"] = \
            input_event.InputEventHandler(
                self.toggle_restrict,
                cmdnames.TOGGLE_RESTRICT_FLAT_REVIEW)

        msg = "FLAT REVIEW PRESENTER: Handlers set up."
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _setup_desktop_bindings(self) -> None:
        """Sets up the flat-review-presenter desktop key bindings."""

        self._desktop_bindings = keybindings.KeyBindings()

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Subtract",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["toggleFlatReviewModeHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Add",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["flatReviewSayAllHandler"],
                2))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Home",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewHomeHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Home",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewPreviousLineHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Up",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewCurrentLineHandler"],
                1))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Up",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewSpellCurrentLineHandler"],
                2))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Up",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewPhoneticCurrentLineHandler"],
                3))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Page_Up",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewNextLineHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Page_Up",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewEndHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Left",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewPreviousItemHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Left",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewAboveHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Begin",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewCurrentItemHandler"],
                1))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Begin",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewSpellCurrentItemHandler"],
                2))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Begin",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewPhoneticCurrentItemHandler"],
                3))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Begin",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewCurrentAccessibleHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Right",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewNextItemHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Right",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewBelowHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_End",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewPreviousCharacterHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_End",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewEndOfLineHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Down",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewCurrentCharacterHandler"],
                1))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Down",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewSpellCurrentCharacterHandler"],
                2))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Down",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewUnicodeCurrentCharacterHandler"],
                3))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "KP_Page_Down",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["reviewNextCharacterHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["showContentsHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["flatReviewCopyHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["flatReviewAppendHandler"]))

        self._desktop_bindings.add(
            keybindings.KeyBinding(
                "",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["flatReviewToggleRestrictHandler"]))

        msg = "FLAT REVIEW PRESENTER: Desktop bindings set up."
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _setup_laptop_bindings(self) -> None:
        """Sets up the flat-review-presenter laptop key bindings."""

        self._laptop_bindings = keybindings.KeyBindings()

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "p",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["toggleFlatReviewModeHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "semicolon",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["flatReviewSayAllHandler"],
                2))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "u",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewPreviousLineHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "u",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_CTRL_MODIFIER_MASK,
                self._handlers["reviewHomeHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "i",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewCurrentLineHandler"],
                1))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "i",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewSpellCurrentLineHandler"],
                2))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "i",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewPhoneticCurrentLineHandler"],
                3))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "o",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewNextLineHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "o",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_CTRL_MODIFIER_MASK,
                self._handlers["reviewEndHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "j",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewPreviousItemHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "j",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_CTRL_MODIFIER_MASK,
                self._handlers["reviewAboveHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "k",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewCurrentItemHandler"],
                1))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "k",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewSpellCurrentItemHandler"],
                2))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "k",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewPhoneticCurrentItemHandler"],
                3))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "k",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_CTRL_MODIFIER_MASK,
                self._handlers["reviewCurrentAccessibleHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "l",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewNextItemHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "l",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_CTRL_MODIFIER_MASK,
                self._handlers["reviewBelowHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "m",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewPreviousCharacterHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "m",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_CTRL_MODIFIER_MASK,
                self._handlers["reviewEndOfLineHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "comma",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewCurrentCharacterHandler"],
                1))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "comma",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewSpellCurrentCharacterHandler"],
                2))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "comma",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewUnicodeCurrentCharacterHandler"],
                3))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "period",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.ORCA_MODIFIER_MASK,
                self._handlers["reviewNextCharacterHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["showContentsHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["flatReviewCopyHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["flatReviewAppendHandler"]))

        self._laptop_bindings.add(
            keybindings.KeyBinding(
                "",
                keybindings.DEFAULT_MODIFIER_MASK,
                keybindings.NO_MODIFIER_MASK,
                self._handlers["flatReviewToggleRestrictHandler"]))

        msg = "FLAT REVIEW PRESENTER: Laptop bindings set up."
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def start(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None
    ) -> None:
        """Starts flat review."""

        if self._context:
            msg = "FLAT REVIEW PRESENTER: Already in flat review"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        msg = "FLAT REVIEW PRESENTER: Starting flat review"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        if script is None:
            script = script_manager.get_manager().get_active_script()
            assert script is not None

        self.get_or_create_context(script)
        if event is None:
            return

        if settings_manager.get_manager().get_setting('speechVerbosityLevel') \
           != settings.VERBOSITY_LEVEL_BRIEF:
            script.present_message(messages.FLAT_REVIEW_START)
        self._item_presentation(script, event)

    def quit(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None
    ) -> None:
        """Quits flat review."""

        if self._context is None:
            msg = "FLAT REVIEW PRESENTER: Not in flat review"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        msg = "FLAT REVIEW PRESENTER: Quitting flat review"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        self._context = None
        focus = focus_manager.get_manager().get_locus_of_focus()
        focus_manager.get_manager().emit_region_changed(focus, mode=focus_manager.FOCUS_TRACKING)
        if event is None or script is None:
            return

        if settings_manager.get_manager().get_setting('speechVerbosityLevel') \
           != settings.VERBOSITY_LEVEL_BRIEF:
            script.present_message(messages.FLAT_REVIEW_STOP)
        script.update_braille(focus)

    @dbus_service.command
    def toggle_flat_review_mode(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = False
    ) -> bool:
        """Toggles between flat review mode and focus tracking mode."""

        tokens = ["FLAT REVIEW PRESENTER: toggle_flat_review_mode. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if self.is_active():
            self.quit(script, event)
            return True

        self.start(script, event)
        return True

    @dbus_service.command
    def go_home(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves to the top left of the current window."""

        tokens = ["FLAT REVIEW PRESENTER: go_home. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._context.go_to_start_of(flat_review.Context.WINDOW)
        self.present_line(script, event)
        return True

    @dbus_service.command
    def go_end(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves to the bottom right of the current window."""

        tokens = ["FLAT REVIEW PRESENTER: go_end. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._context.go_to_end_of(flat_review.Context.WINDOW)
        self.present_line(script, event)
        return True

    @dbus_service.command
    def go_bottom_left(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves to the bottom left of the current window."""

        tokens = ["FLAT REVIEW PRESENTER: go_bottom_left. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._context.go_to_end_of(flat_review.Context.WINDOW)
        self._context.go_to_start_of(flat_review.Context.LINE)
        self.present_line(script, event)
        return True

    @dbus_service.command
    def go_previous_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves to the previous line."""

        tokens = ["FLAT REVIEW PRESENTER: go_previous_line. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_previous_line():
            self.present_line(script, event)
        return True

    @dbus_service.command
    def present_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Presents the current line."""

        tokens = ["FLAT REVIEW PRESENTER: present_line. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._line_presentation(script, event, 1)
        return True

    @dbus_service.command
    def go_next_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves to the next line."""

        tokens = ["FLAT REVIEW PRESENTER: go_next_line. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_next_line():
            self.present_line(script, event)
        return True

    @dbus_service.command
    def spell_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Presents the current line letter by letter."""

        tokens = ["FLAT REVIEW PRESENTER: spell_line. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._line_presentation(script, event, 2)
        return True

    @dbus_service.command
    def phonetic_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Presents the current line letter by letter phonetically."""

        tokens = ["FLAT REVIEW PRESENTER: phonetic_line. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._line_presentation(script, event, 3)
        return True

    @dbus_service.command
    def go_start_of_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves to the beginning of the current line."""

        tokens = ["FLAT REVIEW PRESENTER: go_start_of_line. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._context.go_to_start_of(flat_review.Context.LINE)
        self.present_character(script, event)
        return True

    @dbus_service.command
    def go_end_of_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves to the end of the line."""

        tokens = ["FLAT REVIEW PRESENTER: go_end_of_line. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._context.go_to_end_of(flat_review.Context.LINE)
        self.present_character(script, event)
        return True

    @dbus_service.command
    def go_previous_item(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves to the previous item or word."""

        tokens = ["FLAT REVIEW PRESENTER: go_previous_item. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_previous_word():
            self.present_item(script, event)
        return True

    @dbus_service.command
    def present_item(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Presents the current item/word."""

        tokens = ["FLAT REVIEW PRESENTER: present_item. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._item_presentation(script, event, 1)
        return True

    @dbus_service.command
    def go_next_item(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves to the next item or word."""

        tokens = ["FLAT REVIEW PRESENTER: go_next_item. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_next_word():
            self.present_item(script, event)
        return True

    @dbus_service.command
    def spell_item(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Presents the current item/word letter by letter."""

        tokens = ["FLAT REVIEW PRESENTER: spell_item. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._item_presentation(script, event, 2)
        return True

    @dbus_service.command
    def phonetic_item(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Presents the current word letter by letter phonetically."""

        tokens = ["FLAT REVIEW PRESENTER: phonetic_item. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._item_presentation(script, event, 3)
        return True

    @dbus_service.command
    def go_previous_character(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves to the previous character."""

        tokens = ["FLAT REVIEW PRESENTER: go_previous_character. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_previous_character():
            self.present_character(script, event)
        return True

    @dbus_service.command
    def present_character(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Presents the current character."""

        tokens = ["FLAT REVIEW PRESENTER: present_character. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._character_presentation(script, event, 1)
        return True

    @dbus_service.command
    def go_next_character(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves to the next character."""

        tokens = ["FLAT REVIEW PRESENTER: go_next_character. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_next_character():
            self.present_character(script, event)
        return True

    @dbus_service.command
    def spell_character(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Presents the current character phonetically."""

        tokens = ["FLAT REVIEW PRESENTER: spell_character. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._character_presentation(script, event, 2)
        return True

    @dbus_service.command
    def unicode_current_character(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Presents the current character's unicode value."""

        tokens = ["FLAT REVIEW PRESENTER: unicode_current_character. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._character_presentation(script, event, 3)
        return True

    @dbus_service.command
    def go_above(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves to the character above."""

        tokens = ["FLAT REVIEW PRESENTER: go_above. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_up():
            self.present_item(script, event)
        return True

    @dbus_service.command
    def go_below(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Moves to the character below."""

        tokens = ["FLAT REVIEW PRESENTER: go_below. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_down():
            self.present_item(script, event)
        return True

    @dbus_service.command
    def get_current_object(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> Atspi.Accessible:
        """Returns the current accessible object."""

        tokens = ["FLAT REVIEW PRESENTER: get_current_object. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        return self._context.get_current_object()

    @dbus_service.command
    def present_object(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Presents the current accessible object."""

        tokens = ["FLAT REVIEW PRESENTER: present_object. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if not isinstance(event, input_event.BrailleEvent):
            script.present_object(self._context.get_current_object(), speechonly=True)

        focus_manager.get_manager().emit_region_changed(
            self._context.get_current_object(), mode=focus_manager.FLAT_REVIEW)
        return True

    @dbus_service.command
    def left_click_on_object(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Attempts to synthesize a left click on the current accessible."""

        tokens = ["FLAT REVIEW PRESENTER: left_click_on_object. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        obj = self._context.get_current_object()
        offset = self._context.get_current_text_offset()
        if offset >= 0 and AXEventSynthesizer.click_character(obj, offset, 1):
            return True
        return AXEventSynthesizer.click_object(obj, 1)

    @dbus_service.command
    def right_click_on_object(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Attempts to synthesize a right click on the current accessible."""

        tokens = ["FLAT REVIEW PRESENTER: right_click_on_object. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        obj = self._context.get_current_object()
        offset = self._context.get_current_text_offset()
        if offset >= 0 and AXEventSynthesizer.click_character(obj, offset, 3):
            return True
        return AXEventSynthesizer.click_object(obj, 3)

    @dbus_service.command
    def route_pointer_to_object(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Routes the mouse pointer to the current accessible."""

        tokens = ["FLAT REVIEW PRESENTER: route_pointer_to_object. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        obj = self._context.get_current_object()
        offset = self._context.get_current_text_offset()
        if offset >= 0 and AXEventSynthesizer.route_to_character(obj, offset):
            return True
        return AXEventSynthesizer.route_to_object(obj)

    def _update_braille(self, script: default.Script) -> None:
        """Obtains the braille regions for the current flat review line and displays them."""

        if not settings_manager.get_manager().get_setting("enableBraille") \
           and not settings_manager.get_manager().get_setting("enableBrailleMonitor"):
            debug.print_message(debug.LEVEL_INFO, "BRAILLE: update review disabled", True)
            return

        self._context = self.get_or_create_context(script)
        regions, focused_region = self._context.get_current_braille_regions()
        if not regions:
            braille.refresh(True)
            return

        line = braille.Line()
        line.add_regions(regions)
        braille.set_lines([line])
        braille.setFocus(focused_region)
        braille.refresh(True)

    # TODO - JD: See what adjustments might be needed for the pan_amount parameter
    def pan_braille_left(
        self,
        script: default.Script,
        _event: input_event.InputEvent | None = None,
        pan_amount: int = 0
    ) -> bool:
        """Pans the braille display left."""

        self._context = self.get_or_create_context(script)
        if braille.beginningIsShowing:
            self._context.go_to_start_of(flat_review.Context.LINE)
            self._context.go_previous_character()
            self._update_braille(script)
            return True

        braille.panLeft(pan_amount)
        braille_region, offset_in_zone = braille.getRegionAtCell(1)
        while braille_region and not hasattr(braille_region, "zone"):
            if not braille.panLeft(1):
                break
            braille_region, offset_in_zone = braille.getRegionAtCell(1)

        if braille_region and not hasattr(braille_region, "zone"):
            self._context.go_to_start_of(flat_review.Context.LINE)
            self._context.go_previous_character()
            self._update_braille(script)
            return True

        if braille_region and hasattr(braille_region, "zone"):
            self._context.set_current_zone(braille_region.zone, offset_in_zone)
            self._update_braille(script)
        return True

    # TODO - JD: See what adjustments might be needed for the pan_amount parameter
    def pan_braille_right(
        self,
        script: default.Script,
        _event: input_event.InputEvent | None = None,
        pan_amount: int = 0
    ) -> bool:
        """Pans the braille display right."""

        self._context = self.get_or_create_context(script)
        if braille.endIsShowing:
            self._context.go_next_line()
            self._update_braille(script)
            return True

        braille.panRight(pan_amount)
        braille_region, offset_in_zone = braille.getRegionAtCell(1)
        while braille_region and not hasattr(braille_region, "zone"):
            if not braille.panRight(1):
                break
            braille_region, offset_in_zone = braille.getRegionAtCell(1)

        if braille_region and not hasattr(braille_region, "zone"):
            self._context.go_next_line()
            self._update_braille(script)
            return True

        if braille_region and hasattr(braille_region, "zone"):
            self._context.set_current_zone(braille_region.zone, offset_in_zone)
            self._update_braille(script)
        return True

    def _get_all_lines(
        self,
        script: default.Script,
        _event: input_event.InputEvent | None = None
    ) -> tuple[list[str], tuple[int, int, int, int]]:
        """Returns a (textual lines, current location) tuple."""

        lines = []
        self._context = self.get_or_create_context(script)
        location = self._context.get_current_location()
        self._context.go_to_start_of(flat_review.Context.WINDOW)
        string = self._context.get_current_line_string()
        while string is not None:
            lines.append(string.rstrip("\n"))
            if not self._context.go_next_line():
                break
            string = self._context.get_current_line_string()
        return lines, location

    @dbus_service.command
    def say_all(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Speaks the contents of the entire window."""

        tokens = ["FLAT REVIEW PRESENTER: say_all. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        for string in self._get_all_lines(script, event)[0]:
            if not string.isspace():
                script.speak_message(string, script.speech_generator.voice(string=string))

        return True

    @dbus_service.command
    def show_contents(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Displays the entire flat review contents in a text view."""

        tokens = ["FLAT REVIEW PRESENTER: show_contents. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        lines, location = self._get_all_lines(script, event)
        text = "\n".join(lines)
        title = guilabels.FLAT_REVIEW_CONTENTS
        self._gui = FlatReviewContextGUI(script, title, text, location)
        self._gui.show_gui()
        return True

    @dbus_service.command
    def copy_to_clipboard(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Copies the string just presented to the clipboard."""

        tokens = ["FLAT REVIEW PRESENTER: copy_to_clipboard. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not self.is_active():
            if notify_user:
                script.present_message(messages.FLAT_REVIEW_NOT_IN)
            return True

        script.get_clipboard_presenter().set_text(self._current_contents.rstrip("\n"))
        if notify_user:
            script.present_message(messages.FLAT_REVIEW_COPIED)
        return True

    @dbus_service.command
    def append_to_clipboard(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """Appends the string just presented to the clipboard."""

        tokens = ["FLAT REVIEW PRESENTER: append_to_clipboard. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not self.is_active():
            if notify_user:
                script.present_message(messages.FLAT_REVIEW_NOT_IN)
            return True

        script.get_clipboard_presenter().append_text(self._current_contents.rstrip("\n"))
        if notify_user:
            script.present_message(messages.FLAT_REVIEW_APPENDED)
        return True

    @dbus_service.getter
    def get_is_restricted(self) -> bool:
        """Returns whether flat review is restricted to the current object."""

        return settings_manager.get_manager().get_setting("flatReviewIsRestricted")

    @dbus_service.setter
    def set_is_restricted(self, value: bool) -> bool:
        """Sets whether flat review is restricted to the current object."""

        msg = f"FLAT REVIEW PRESENTER: Setting is-restricted to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        settings_manager.get_manager().set_setting("flatReviewIsRestricted", value)
        return True

    @dbus_service.command
    def toggle_restrict(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True
    ) -> bool:
        """ Toggles the restricting of flat review to the current object. """

        tokens = ["FLAT REVIEW PRESENTER: toggle_restrict. Script:", script,
                  "Event:", event, "notify_user:", notify_user]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._restrict = not self._restrict
        settings_manager.get_manager().set_setting("flatReviewIsRestricted", self._restrict)

        if self._restrict:
            if notify_user:
                script.present_message(messages.FLAT_REVIEW_RESTRICTED)
        else:
            if notify_user:
                script.present_message(messages.FLAT_REVIEW_UNRESTRICTED)
        if self.is_active():
            # Reset the context
            self._context = None
            self.start()

        return True

    def _line_presentation(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        speech_type: int = 1
    ) -> bool:
        """Presents the current line."""

        self._context = self.get_or_create_context(script)
        line_string = self._context.get_current_line_string()
        voice = script.speech_generator.voice(string=line_string)

        if not isinstance(event, input_event.BrailleEvent):
            if not line_string or line_string == "\n":
                script.speak_message(messages.BLANK)
            elif line_string.isspace():
                script.speak_message(messages.WHITE_SPACE)
            elif line_string.isupper() and (speech_type < 2 or speech_type > 3):
                script.speak_message(line_string, voice)
            elif speech_type == 2:
                script.spell_item(line_string)
            elif speech_type == 3:
                script.spell_phonetically(line_string)
            else:
                manager = speech_and_verbosity_manager.get_manager()
                line_string = manager.adjust_for_presentation(
                    self._context.get_current_object(), line_string)
                script.speak_message(line_string, voice)

        focus_manager.get_manager().emit_region_changed(
            self._context.get_current_object(), mode=focus_manager.FLAT_REVIEW)
        self._update_braille(script)
        self._current_contents = line_string
        return True

    def _item_presentation(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        speech_type: int = 1
    ) -> bool:
        """Presents the current item/word."""

        self._context = self.get_or_create_context(script)
        word_string = self._context.get_current_word_string()
        voice = script.speech_generator.voice(string=word_string)
        if not isinstance(event, input_event.BrailleEvent):
            if not word_string or word_string == "\n":
                script.speak_message(messages.BLANK)
            else:
                line_string = self._context.get_current_line_string()
                if line_string == "\n":
                    script.speak_message(messages.BLANK)
                elif word_string.isspace():
                    script.speak_message(messages.WHITE_SPACE)
                elif word_string.isupper() and speech_type == 1:
                    script.speak_message(word_string, voice)
                elif speech_type == 2:
                    script.spell_item(word_string)
                elif speech_type == 3:
                    script.spell_phonetically(word_string)
                elif speech_type == 1:
                    manager = speech_and_verbosity_manager.get_manager()
                    word_string = manager.adjust_for_presentation(
                        self._context.get_current_object(), word_string)
                    script.speak_message(word_string, voice)

        focus_manager.get_manager().emit_region_changed(
            self._context.get_current_object(), mode=focus_manager.FLAT_REVIEW)
        self._update_braille(script)
        self._current_contents = word_string
        return True

    def _character_presentation(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        speech_type: int = 1
    ) -> bool:
        """Presents the current character."""

        focus = focus_manager.get_manager().get_locus_of_focus()
        if not self._context and AXObject.supports_text(focus):
            char_string = AXText.get_character_at_offset(focus)[0]
        else:
            self._context = self.get_or_create_context(script)
            char_string = self._context.get_current_character_string()
        if not isinstance(event, input_event.BrailleEvent):
            if not char_string:
                script.speak_message(messages.BLANK)
            else:
                if char_string == "\n" and speech_type != 3:
                    script.speak_message(messages.BLANK)
                elif speech_type == 3:
                    script.speak_message(messages.UNICODE % f"{ord(char_string):04x}")
                elif speech_type == 2:
                    script.spell_phonetically(char_string)
                else:
                    script.speak_character(char_string)

        if not self._context:
            return True

        focus_manager.get_manager().emit_region_changed(
            self._context.get_current_object(), mode=focus_manager.FLAT_REVIEW)
        self._update_braille(script)
        self._current_contents = char_string
        return True

class FlatReviewContextGUI:
    """Presents the entire flat review context in a text view"""

    def __init__(
        self,
        script: default.Script,
        title: str,
        text: str,
        location: tuple[int, int, int, int]
    ) -> None:
        self._script: default.Script = script
        self._gui: Gtk.Dialog = self._create_dialog(title, text, location)

    def _create_dialog(
        self,
        title: str,
        text: str,
        location: tuple[int, int, int, int]
    ) -> Gtk.Dialog:
        """Creates the dialog."""

        dialog = Gtk.Dialog(title,
                            None,
                            Gtk.DialogFlags.MODAL,
                            (Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE))
        dialog.set_default_size(800, 600)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.set_hexpand(True)
        scrolled_window.set_vexpand(True)

        textbuffer = Gtk.TextBuffer()
        textbuffer.set_text(text)

        line_index, _zone_index, word_index, char_index = location
        start_iter = textbuffer.get_start_iter()
        start_iter.forward_lines(line_index)
        for _ in range(word_index):
            start_iter.forward_word_end()
            # This is needed to get past any punctuation. Note that we cannot check starts_word
            # here. Example: The letter after an apostrophe is reported as a word start.
            while not start_iter.get_char().isspace():
                if not start_iter.forward_char():
                    break

        # The text iter word can start with one or more spaces. Move to the beginning of the flat
        # review word before advancing by character.
        while not start_iter.starts_word():
            if not start_iter.forward_char():
                break

        start_iter.forward_chars(char_index)
        textbuffer.place_cursor(start_iter)

        textview = Gtk.TextView(buffer=textbuffer)
        textview.set_wrap_mode(Gtk.WrapMode.WORD)
        scrolled_window.add(textview) # pylint: disable=no-member
        dialog.get_content_area().pack_start(scrolled_window, True, True, 0)
        dialog.connect("response", self.on_response)
        return dialog

    def on_response(self, _dialog: Gtk.Dialog, response: Gtk.ResponseType) -> None:
        """Handler for the 'response' signal of the dialog."""

        if response == Gtk.ResponseType.CLOSE:
            self._gui.destroy()

    def show_gui(self) -> None:
        """Shows the dialog."""

        self._gui.show_all() # pylint: disable=no-member
        self._gui.present_with_time(time.time())


_presenter: FlatReviewPresenter = FlatReviewPresenter()

def get_presenter() -> FlatReviewPresenter:
    """Returns the Flat Review Presenter"""

    return _presenter
